/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2012 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: MergeObject.java 6254 2012-05-02 21:35:26Z cazauxj $
 * --------------------------------------------------------------------------
 */
package org.ow2.util.merge.object;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

/**
 * Merge 2 Java objects
 * @author Jeremy Cazaux
 */
public class MergeObject {

    /**
     * The logger
     */
    private static final Log logger = LogFactory.getLog(MergeObject.class);

    /**
     * Java lang pattern
     */
    private static final Pattern pattern = Pattern.compile("java.lang.+");

    /**
     * Merge two Java Objects
     * Merge the object two into the object one
     * @param o1 The object one
     * @param o2 The object two
     */
    public static void mergeObject(Object o1, final Object o2) {
        mergeObject(o1, o2, null);
    }

    /**
     * Merge two Java Objects
     * Merge the object two into the object one
     * @param o1 The object one
     * @param o2 The object two
     * @param classIdFieldMap map between the {@link Class} of an {@link Object} and its id field
     *                        (use to improve the merge of Collection)
     */
    public static void mergeObject(Object o1, final Object o2, final Map<Class, Field> classIdFieldMap) {
        if (o1 == null) {
            o1 = o2;
        } else {
            if (o2 != null) {
                if (!o1.getClass().getName().equals(o2.getClass().getName())) {
                    throw new IllegalStateException("Cannot merge two object of different class.");
                }

                List<Field> fields = getAllField(o1.getClass());
                if (fields != null) {
                    for (Field field: fields) {
                        Object v1 = getValue(field, o1);
                        Object v2 = getValue(field, o2);
                        int modifier = field.getModifiers();
                        if (!Modifier.isFinal(modifier) && !Modifier.isStatic(modifier)) {
                            if (v1 == null) {
                                setValue(field, o1, v2);
                            } else {
                                if (v1 instanceof Collection) {
                                    Collection collection1 = Collection.class.cast(v1);
                                    Collection collection2 = Collection.class.cast(v2);
                                    for (Object item: collection2) {
                                        Object matchingObject = getMatchingObject(collection1, item, classIdFieldMap);
                                        if (matchingObject != null) {
                                            mergeObject(matchingObject, item, classIdFieldMap);
                                        } else {
                                            collection1.add(item);
                                        }

                                    }
                                } else if (v1 instanceof Map) {
                                    Map map1 = Map.class.cast(v1);
                                    Map map2 = Map.class.cast(v2);
                                    if (map2 != null) {
                                        for (Object key: map2.keySet()) {
                                            if (map1.containsKey(key)) {
                                                mergeObject(map1.get(key), map2.get(key), classIdFieldMap);
                                            } else {
                                                map1.put(key, map2.get(key));
                                            }
                                        }
                                    }
                                } else {
                                    boolean javaLangClass = pattern.matcher(field.getType().getName()).matches();
                                    if (v2 != null && !v2.getClass().isPrimitive() && !javaLangClass) {
                                        mergeObject(v1, v2, classIdFieldMap);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * @param collection A @{link Collection}
     * @param object An {@link Object}
     * @param classIdFieldMap map between the {@link Class} of an {@link Object} and its id field
     *                    (use to merge Collection)
     * @return the matching object if found. Otherwise, return null.
     */
    private static Object getMatchingObject(final Collection collection, final Object object,
                                            final Map<Class, Field> classIdFieldMap) {
        for (Object item: collection) {
            if (item.equals(object)) {
                return item;
            } else if (item.getClass().equals(object.getClass())) {
                //get fields of the class
                Field[] fields = item.getClass().getDeclaredFields();
                if (fields != null) {
                    List<Field> fieldList = Arrays.asList(fields);

                    //get the id field of an Object
                    if (classIdFieldMap != null) {
                        Field idField = classIdFieldMap.get(item.getClass());
                        if (idField != null && fieldList.contains(idField)) {
                            Object v1 = getValue(idField, item);
                            Object v2 = getValue(idField, object);
                            //test if id field of objects are equals
                            if (v1 != null && v2 != null && v1.equals(v2)) {
                                return item;
                            }
                        }
                    }
                }
            }
        }
        return null;
    }

    /**
     * @param clazz A {@link Class}
     * @return all fields (public, protected, privated, inherit) of a a {@link Class}
     */
    public static List<Field> getAllField(final Class clazz) {
        List<Field> fields = new ArrayList<Field>();
        if (clazz != null) {
            fields.addAll(Arrays.asList(clazz.getDeclaredFields()));
            List<Field> fieldList = getAllField(clazz.getSuperclass());
            if (fieldList != null) {
                fields.addAll(fieldList);
            }
        }
        return fields;
    }

    /**
     * @param field A {@link Field}
     * @param object An {@link Object}
     * @return the value of the given field of the given object
     */
    public static synchronized Object getValue(final Field field, final Object object) {
        boolean isPrivateField = !field.isAccessible();
        Object value = null;
        if (isPrivateField) {
            field.setAccessible(true);
        }
        try {
            value = field.get(object);
        } catch (IllegalAccessException e) {
            logger.error("Cannot get value of field " + field.getName() + " of the object " + object.getClass().getName(), e);
        }
        if (isPrivateField) {
            field.setAccessible(false);
        }
        return value;
    }

    /**
     *
     * @param field A {@link Field}
     * @param object The {@link Object} whose field should be modified
     * @param value The new value of the field to set
     */
    public static synchronized void setValue(final Field field, final Object object, final Object value) {
        boolean isPrivateField = !field.isAccessible();
        if (isPrivateField) {
            field.setAccessible(true);
        }
        try {
            field.set(object, value);
        } catch (IllegalAccessException e) {
            logger.error("Cannot set field " + field.getName() + " of the object " + object.getClass().getName() + " to " + value, e);
        }
        if (isPrivateField) {
            field.setAccessible(false);
        }
    }
}
